dict_rhythms = {
  "E(2,3)" : "A common Afro-Cuban drum pattern when started on the second onset as in [101]. For example, it is the conga rhythm of the (6/8)-time Swing Tumbao. It is common in Latin American music, as for example in the Cueca, and the coros de clave. It is common in Arabic music, as for example in the Al Taer rhythm of Nubia. It is also a rhythmic pattern of the Drum Dance of the Slavey Indians of Northern Canada",
  "E(2,5)" : "A rhythm found in Greece, Namibia, Rwanda and Central Africa. It is also a 13th century Persian rhythm called Khafif-e-ramal, as well as the rhythm of the dance Makedonka from the FYROM. Tchaikovsky used it as the metric pattern in the second movement of his Symphony No. 6. Started on the second onset as in [10010] it is a rhythm found in Central Africa, Bulgaria, Turkey, Turkestan and Norway. It is also the metric pattern of Dave Brubeck's Take Five, as well as Mars from The Planets by Gustav Holst. Rotated as in [10010100100], it is a Serbian rhythmic pattern. When it is started on the fourth (last) onset it is the Daasa al kbiri rhythmic pattern of Yemen.",
  "E(3,4)" : "It is the archetypal pattern of the Cumbia from Colombia, as well as a Calypso rhythm from Trinidad. It is also a thirteenth century Persian rhythm called Khalif-e-saghil, as well as the trochoid choreic rhythmic pattern of ancient Greece.",
  "E(3,5)" : "When started on the second onset, it is a thirteenth century Persian rhythm by the name of Khafif-e-ramal, as well as a Rumanian folk-dance rhythm.",
  "E(3,7)" : "It is the Ruchenitza rhythm used in a Bulgarian folk-dance. It is also the metric pattern of Pink Floyd's Money.",
  "E(3,8)" : "It is none other than one of the most famous rythms on the planet. In Cuba it goes by the name of the tresillo and in the USA is often called the Habanera rhythm used in hundreds of rockabilly songs during the 1950s. It can often be heard in early rock-and-roll hits in the left-hand patterns of the piano, or played on the string bass or saxophone. A good example is the bass rhythm in Elvis Presley's Hound Dog. The tresillo pattern is also found widely in West African traditional music. For example, it is played on the atoke bell in the Sohu, an Ewe dance from Ghana. The tresillo can also be recognized as the first bar of the ubiquitous two-bar clave Son given by [1001001000101000].",
  "E(4,7)" : "It is a Ruchenitza Bulgarian folk-dance rhythm.",
  "E(4,9)" : "It is the Aksak rhythm of Turkey. It is also the metric pattern used by Dave Brubeck in his piece Rondo a la Turk.",
  "E(4,11)" : "It is the metric pattern used by Frank Zappa in his piece titled Outside Now.",
  "E(4,12)" : "It is periodic with four repetitions of E(1,3) = [100]. It is the (12/8)-time Fandago clapping pattern in the Flamenco music of southern Spain, where 1 denotes a loud clap and 0 soft clap.",
  "E(4,15)" : "It is the metric pattern of the pancam savari tal of North Indian music.",
  "E(5,6)" : "It yields the York-Samai pattern, a popular Arabic rhythm. It is also a handclapping rhythm used in the Al Medemi songs of Oman.",
  "E(5,7)" : "It is the Nawakhat pattern, a popular Arabic rhythm. In Nubia it is called the Al Noht rhythm.",
  "E(5,8)" : "It is the Cuban cinquillo pattern, the Malfuf rhythmic pattern of Egypt, as well as the Korean Nong Pyon drum pattern. Started on the second onset, it is a popular Middle Eastern rhythm, as well as the Timini rhythm of Senegal, the Adzogbo dance rhythm of Benin, the Spanish Tango, the Maksum of Egypt, and a 13th century Persian rhythm, the Al-saghil-alsani. When it is started on the third onset it is the Musemmen rhythm of Turkey. When it is started on the fourth onset it is the Kromanti rhythm of Surinam.",
  "E(5,9)" : "It is a popular Arabic rhythm called Agsag-Samai. Started on the second onset, it is a drum pattern used by the Venda in South Africa, as well as a Rumanian folk-dance rhythm. It is also the rhythmic pattern of the Sigaktistos rhythm of Greece, and the Samai aktsak rhythm of Turkey. Started on the third onset, it is the rhythmic pattern of the Nawahiid rhythm of Turkey.",
  "E(5,11)" : "It is the metric pattern of the Savari tala used in the Hindustani music of India. It is also a rhythmic pattern used in Bulgaria and Serbia. In Bulgaria is is used in the Kopanitsa. This metric pattern has been used by Moussorgsky in Pictures at an Exhibition. Started on the third onset, it is the rhythm of the dance Kalajdzijsko Oro from the FYROM, and it appears in Bulgarian music as well.",
  "E(5,12)" : "It is a common rhythm played in the Central African Republic by the Aka Pygmies. It is also the Venda clapping pattern of a South African children's song, and a rhythm pattern used in the FYROM. Started on the second onset, it is the Columbia bell pattern popular in Cuba and West Africa, as well as a drumming pattern used in the Chakacha dance of Kenya, and also used in the FYROM . Started on the third onset, it is the Bemba bell pattern used in Northern Zimbabwe, and the rhythm of the dance from the FYROM Ibraim Odza Oro. Started on the fourth onset, it is the Fume Fume bell pattern popular in West Africa, and is a rhythm used in the former Yugoslavia. Finally, when started on the fifth onset it is the Salve bell pattern used in the Dominican Republic in a rhythm called Canto de Vela in honor of the Virgin Mary, as well as the drum rhythmic pattern of the Moroccan Al Kudam.",
  "E(5,13)" : "It is a rhythm from the FYROM which is also played by starting it on the fourth onset as follows: [1010010010100].",
  "E(5,16)" : "It is the Bossa-Nova rhythm necklace of Brazil. The actual Bossa-Nova rhythm usually starts on the third onset as follows: [1001001000100100]. However, other starting places are also documented in world music practices, such as [1001001001000100].",
  "E(6,7)" : "It is the Pontakos rhythm of Greece when started on the sixth (last) onset.",
  "E(6,13)" : "It is the rhythm of the dance Mama Cone pita from the FYROM. Started on the third onset, it is the rhythm of the dance Postupano Oro from the FYROM, as well as the Krivo Plovdivsko Horo of Bulgaria.",
  "E(7,8)" : "When started on the seventh (last) onset, is a typical rhythm played on the Bendir (frame drum), and used in the accompaniment of songs of the Tuareg people of Libya.",
  "E(7,9)" : "It is the Bazaragana rhythmic pattern of Greece.",
  "E(7,10)" : "It is the Lenk fahhte rhythmic pattern of Turkey.",
  "E(7,12)" : "It is a common West African bell pattern. For example, it is used in the Mpre rhythm of the Ashanti people of Ghana. Started on the seventh (last) onset, it is a Yoruba bell pattern of Nigeria, a Babenzele pattern of Central Africa, and a Mende pattern of Sierra Leone.",
  "E(7,15)" : "It is a Bulgarian rhythm when started on the third onset.",
  "E(7,16)" : "It is a Samba rhythm necklace from Brazil. The actual Samba rhythm is [1010010101001010] obtained by starting E(7, 16) on the last onset, and it coincides with a rhythm from the FYROM. When E(7, 16) is started on the fifth onset it is a clapping pattern from Ghana. When it is started on the second onset it is a rhythmic pattern found in the former Yugoslavia.",
  "E(7,17)" : "It is a rhythm from the FYROM when started on the second onset.",
  "E(7,18)" : "It is a Bulgarian rhythmic pattern.",
  "E(8,17)" : "It is a Bulgarian rhythmic pattern which is also started on the fifth onset.",
  "E(8,19)" : "It is a Bulgarian rhythmic pattern when started on the second onset.",
  "E(9,14)" : "When started on the second onset, it is the rhythmic pattern of the Tsofyan rhythm of Algeria.",
  "E(9,16)" : "It is a rhythm necklace used in the Central African Republic. When it is started on the second onset it is a bell pattern of the Luba people of Congo. When it is started on the fourth onset it is a rhythm played in West and Central Africa, as well as a cow-bell pattern in the Brazilian samba. When it is started on the penultimate onset it is the bell pattern of the Ngbaka-Maibo rhythms of the Central African Republic.",
  "E(9,22)" : "It is a Bulgarian rhythmic pattern when started on the second onset.",
  "E(9,23)" : "It is a Bulgarian rhythm.",
  "E(11,12)" : "When started on the second onset, is thedrum pattern of the Rahmani (a cylindrical double-headed drum) used in the Sot silam dance from Mirbat in the South of Oman.",
  "E(11,24)" : "It is a rhythm necklace of the Aka Pygmies of Central Africa. It is usually started on the seventh onset. Started on the second onset, it is a Bulgarian rhythm.",
  "E(13,24)" : "It is another rhythm necklace of the Aka Pygmies of the upper Sangha. Started on the penultimate onset, it is the Bobangi metal-blade pattern used by the Aka Pygmies.",
  "E(15,34)" : "It is a Bulgarian rhythmic pattern when started on the penultimate onset."
}